import numpy as np

from caliber.binary_classification.metrics.bias import model_bias


class ModelBiasBinaryClassificationConstantShift:
    def __init__(self, step_size: float = 1.0):
        self.step_size = step_size
        self._shift = None

    def fit(self, probs: np.ndarray, targets: np.ndarray) -> None:
        self._shift = model_bias(targets, probs)

    def predict_proba(self, probs: np.ndarray) -> np.ndarray:
        return np.clip(probs + self.step_size * self._shift, 0, 1)

    def predict(self, probs: np.ndarray) -> np.ndarray:
        return (self.predict_proba(probs) >= 0.5).astype(int)



from typing import Callable, Optional

import numpy as np

from caliber.binary_classification.linear_scaling.base import (
    CustomBinaryClassificationLinearScaling,
)
from caliber.binary_classification.linear_scaling.brute_fit_mixin_linear_scaling import (
    BinaryClassificationLinearScalingBruteFitMixin,
)


class CustomPerformanceBinaryClassificationLinearScaling(
    BinaryClassificationLinearScalingBruteFitMixin,
    CustomBinaryClassificationLinearScaling,
):
    def __init__(
        self,
        loss_fn: Callable[[np.ndarray, np.ndarray], float],
        threshold: float,
        minimize_options: Optional[dict] = None,
        has_intercept: bool = True,
    ):
        super().__init__(loss_fn, minimize_options, has_intercept)
        self._threshold = threshold

    def _get_output_for_loss(self, params: np.ndarray, probs: np.ndarray) -> np.ndarray:
        return (self._predict_proba(params, probs) >= self._threshold).astype(int)

    @property
    def threshold(self):
        return self._threshold

    def predict(self, probs: np.ndarray) -> np.ndarray:
        return (self.predict_proba(probs) >= self._threshold).astype(int)



from caliber.binary_classification.binning.histogram_binning.base import (
    HistogramBinningBinaryClassificationModel,
)
from caliber.binary_classification.binning.isotonic_regression.base import (
    IsotonicRegressionBinaryClassificationModel,
)
from caliber.binary_classification.binning.iterative_binning.base import (
    IterativeBinningBinaryClassificationModel,
)
from caliber.binary_classification.constant_shift.model_bias.base import (
    ModelBiasBinaryClassificationConstantShift,
)
from caliber.binary_classification.linear_scaling.calibration.asce_linear_scaling import (
    ASCEBinaryClassificationLinearScaling,
)
from caliber.binary_classification.linear_scaling.calibration.brier_linear_scaling import (
    BrierBinaryClassificationLinearScaling,
)
from caliber.binary_classification.linear_scaling.calibration.cross_entropy_linear_scaling import (
    CrossEntropyBinaryClassificationLinearScaling,
)
from caliber.binary_classification.linear_scaling.calibration.ece_linear_scaling import (
    ECEBinaryClassificationLinearScaling,
)
from caliber.binary_classification.linear_scaling.performance.bal_acc_linear_scaling import (
    BalancedAccuracyBinaryClassificationLinearScaling,
)
from caliber.binary_classification.linear_scaling.performance.f1_linear_scaling import (
    NegativeF1BinaryClassificationLinearScaling,
    PositiveF1BinaryClassificationLinearScaling,
)
from caliber.binary_classification.linear_scaling.performance.positive_negative_rates_linear_scaling import (
    PositiveNegativeRatesBinaryClassificationLinearScaling,
)
from caliber.binary_classification.linear_scaling.performance.precision_fixed_recall import (
    PrecisionFixedRecallBinaryClassificationLinearScaling,
)
from caliber.binary_classification.linear_scaling.performance.predictive_values_linear_scaling import (
    PredictiveValuesBinaryClassificationLinearScaling,
)
from caliber.binary_classification.linear_scaling.performance.recall_fixed_precision import (
    RecallFixedPrecisionBinaryClassificationLinearScaling,
)
from caliber.binary_classification.linear_scaling.performance.righteousness_linear_scaling import (
    RighteousnessBinaryClassificationLinearScaling,
)



import numpy as np
from sklearn.calibration import calibration_curve


def average_squared_calibration_error(
    targets: np.ndarray, probs: np.ndarray, n_bins: int = 10
) -> float:
    binned_ey, binned_probs = calibration_curve(
        targets, probs, n_bins=n_bins, strategy="quantile"
    )
    return float(np.mean((binned_ey - binned_probs) ** 2))



import abc
from typing import Callable, Optional

import numpy as np

from caliber.binary_classification.linear_scaling.base import (
    CustomBinaryClassificationLinearScaling,
)


class CustomCalibrationBinaryClassificationLinearScaling(
    CustomBinaryClassificationLinearScaling, abc.ABC
):
    def __init__(
        self,
        loss_fn: Callable[[np.ndarray, np.ndarray], float],
        minimize_options: Optional[dict] = None,
        has_intercept: bool = True,
    ):
        super().__init__(loss_fn, minimize_options, has_intercept)

    def _get_output_for_loss(self, params: np.ndarray, probs: np.ndarray) -> np.ndarray:
        return self._predict_proba(params, probs)

    def predict(self, probs: np.ndarray) -> np.ndarray:
        return (self.predict_proba(probs) >= 0.5).astype(int)



import numpy as np

from caliber.binary_classification.base_smooth_fit_mixin import (
    BinaryClassificationSmoothFitMixin,
)


class BinaryClassificationLinearScalingSmoothFitMixin(
    BinaryClassificationSmoothFitMixin
):
    def _get_x0(self) -> np.ndarray:
        return np.array([0.0, 1.0]) if self._has_intercept else np.array(1.0)



import abc
from copy import deepcopy
from typing import Any, Tuple

import numpy as np


class BinaryClassificationIterativeModel:
    def __init__(self, model: Any, n_rounds: int, seed: int = 0):
        self._model = model
        self.n_rounds = n_rounds
        self._models = []
        self._rng = np.random.default_rng(seed)

    def fit(self, features: np.ndarray, targets: np.ndarray):
        self._models = []
        model = deepcopy(self._model)
        for t in range(self.n_rounds):
            self._fit(model, features, targets)
            self._models.append(model)
            model, features, targets = self._update(model, features, targets)

    @abc.abstractmethod
    def _fit(self, model: Any, features: np.ndarray, targets: np.ndarray):
        pass

    @abc.abstractmethod
    def _predict_proba(self, model: Any, features: np.ndarray):
        pass

    @abc.abstractmethod
    def _update(
        self, model: Any, features: np.ndarray, targets: np.ndarray
    ) -> Tuple[Any, np.ndarray, np.ndarray]:
        pass

    @abc.abstractmethod
    def _predict(self, model: Any, features: np.ndarray) -> np.ndarray:
        pass



import abc
from typing import Callable, Optional

import numpy as np


class CustomBinaryClassificationModel:
    def __init__(
        self,
        loss_fn: Callable[[np.ndarray, np.ndarray], float],
        minimize_options: Optional[dict] = None,
    ):
        self._loss_fn = loss_fn
        self._params = None
        self._minimize_options = self._config_minimize_options(minimize_options)

    @abc.abstractmethod
    def fit(self, probs: np.ndarray, targets: np.ndarray):
        pass

    def predict_proba(self, probs: np.ndarray) -> np.ndarray:
        return self._predict_proba(self._params, probs)

    @staticmethod
    @abc.abstractmethod
    def _predict_proba(params: np.ndarray, probs: np.ndarray) -> np.ndarray:
        pass

    @abc.abstractmethod
    def predict(self, probs: np.ndarray):
        self._check_probs(probs)
        pass

    @abc.abstractmethod
    def _get_output_for_loss(self, params: np.ndarray, probs: np.ndarray) -> np.ndarray:
        pass

    @staticmethod
    def _check_targets(targets: np.ndarray) -> None:
        if set(targets) != {0, 1}:
            raise ValueError(
                "`targets` must must include only values in {0, 1}. Both classes must be present."
            )
        if targets.ndim != 1:
            raise ValueError("`targets` must be a one-dimensional array.")

    @staticmethod
    def _check_probs(probs: np.ndarray) -> None:
        if probs.ndim != 1:
            raise ValueError("`probs` must be a one-dimensional array.")
        if np.sum(probs > 1) or np.sum(probs < 0):
            raise ValueError("All values in `probs` must be between 0 and 1.")

    @abc.abstractmethod
    def _config_minimize_options(self, minimize_options: Optional[dict]) -> dict:
        pass



from typing import Optional

from caliber.binary_classification.linear_scaling.brute_fit_mixin_linear_scaling import (
    BinaryClassificationLinearScalingBruteFitMixin,
)
from caliber.binary_classification.linear_scaling.calibration.base import (
    CustomCalibrationBinaryClassificationLinearScaling,
)
from caliber.binary_classification.metrics.asce import average_squared_calibration_error


class ASCEBinaryClassificationLinearScaling(
    BinaryClassificationLinearScalingBruteFitMixin,
    CustomCalibrationBinaryClassificationLinearScaling,
):
    def __init__(
        self, minimize_options: Optional[dict] = None, has_intercept: bool = True
    ):
        super().__init__(
            loss_fn=average_squared_calibration_error,
            minimize_options=minimize_options,
            has_intercept=has_intercept,
        )



from caliber.binary_classification.metrics.asce import average_squared_calibration_error
from caliber.binary_classification.metrics.bias import (
    absolute_model_bias,
    squared_model_bias,
)
from caliber.binary_classification.metrics.ece import expected_calibration_error



from typing import List, Tuple

from caliber.binary_classification.base_brute_fit_mixin import (
    BinaryClassificationBruteFitMixin,
)


class BinaryClassificationLinearScalingBruteFitMixin(BinaryClassificationBruteFitMixin):
    def _get_ranges(self) -> List[Tuple]:
        return [(-2, 2), (0, 4)] if self._has_intercept else [(0, 4)]

    @staticmethod
    def _get_Ns() -> int:
        return 200



import abc
from typing import Callable, Optional

import numpy as np
from scipy.special import expit, logit

from caliber.binary_classification.base import CustomBinaryClassificationModel


class CustomBinaryClassificationLinearScaling(CustomBinaryClassificationModel, abc.ABC):
    def __init__(
        self,
        loss_fn: Callable[[np.ndarray, np.ndarray], float],
        minimize_options: Optional[dict] = None,
        has_intercept: bool = True,
    ):
        self._has_intercept = has_intercept
        super().__init__(loss_fn, minimize_options)

    def _predict_proba(self, params: np.ndarray, probs: np.ndarray) -> np.ndarray:
        return (
            expit(params[0] + params[1] * logit(probs))
            if self._has_intercept
            else expit(params * logit(probs))
        )
